// file      : xsd/cxx/tree/elements.hxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#ifndef CXX_TREE_ELEMENTS_HXX
#define CXX_TREE_ELEMENTS_HXX

#include <sstream>

#include <cult/containers/deque.hxx>

#include <cxx/elements.hxx>

#include <cxx/tree/cli.hxx>


namespace CXX
{
  namespace Tree
  {
    struct Counts
    {
      Counts ()
          : global_types (0),
            global_elements (0),
            generated_global_elements (0)
      {
      }

      UnsignedLong global_types;
      UnsignedLong global_elements;
      UnsignedLong generated_global_elements;
    };

    //
    //
    class Context: public CXX::Context
    {
    public:
      Context (std::wostream& o,
               CLI::Options const& ops,
               Counts const& counts_)
          : CXX::Context (o,
                          ops.value<CLI::char_type> (),
                          ops.value<CLI::include_with_brackets> (),
                          ops.value<CLI::include_prefix> (),
                          ops.value<CLI::export_symbol> (),
                          ops.value<CLI::namespace_map> (),
                          ops.value<CLI::namespace_regex> (),
                          ops.value<CLI::namespace_regex_trace> (),
                          ops.value<CLI::generate_inline> ()),
            options (options_),
            counts (counts_),
            parser_type (parser_type_),
            xml_element_type (xml_element_type_),
            xml_attribute_type (xml_attribute_type_),
            options_ (ops),
            parser_type_ (L"::xsd::cxx::xml::dom::parser< " + char_type + L" >"),
            xml_element_type_ (L"::xsd::cxx::xml::dom::element< " + char_type + L" >"),
            xml_attribute_type_ (L"::xsd::cxx::xml::dom::attribute< " + char_type + L" >"),
            ns_scope (ns_scope_),
            scope (scope_),
            inner_name (inner_name_),
            ns_scope_stack (ns_scope_stack_),
            scope_stack (scope_stack_)
      {
      }

    protected:
      Context (Context& c)
          : CXX::Context (c),
            options (c.options),
            counts (c.counts),
            parser_type (c.parser_type),
            xml_element_type (c.xml_element_type),
            xml_attribute_type (c.xml_attribute_type),
            ns_scope (c.ns_scope),
            scope (c.scope),
            inner_name (c.inner_name),
            ns_scope_stack (c.ns_scope_stack),
            scope_stack (c.scope_stack)
      {
      }

      Context (Context& c, std::wostream& o)
          : CXX::Context (c, o),
            options (c.options),
            counts (c.counts),
            parser_type (c.parser_type),
            xml_element_type (c.xml_element_type),
            xml_attribute_type (c.xml_attribute_type),
            ns_scope (c.ns_scope),
            scope (c.scope),
            inner_name (c.inner_name),
            ns_scope_stack (c.ns_scope_stack),
            scope_stack (c.scope_stack)
      {
      }

    public:
      // Return infoscope name.
      //
      String
      iscope_name (String const& scope_name, String const& name)
      {
        return scope_name + L"::" + unclash (scope_name, name);
      }

    public:
      Void
      enter_ns_scope (String const& name)
      {
        ns_scope_stack.push_back (name);
        update_ns_scope ();
      }

      Void
      leave_ns_scope ()
      {
        ns_scope_stack.pop_back ();
        update_ns_scope ();
      }

    public:
      Void
      enter_scope (String const& name)
      {
        scope_stack.push_back (name);
        update_scope ();
      }

      Void
      leave_scope ()
      {
        scope_stack.pop_back ();
        update_scope ();
      }

  private:
      Void
      update_ns_scope ()
      {
        ns_scope.clear ();

        Boolean first (true);

        for (NamespaceStack::Iterator i (ns_scope_stack.begin ());
             i != ns_scope_stack.end ();
             ++i)
        {
          // We only qualify names until the namespace level.
          //
          if (first)
            first = false;
          else
            ns_scope += L"::";

          ns_scope += *i;
        }
      }

      Void
      update_scope ()
      {
        scope.clear ();

        Boolean first (true);

        for (ScopeStack::Iterator i (scope_stack.begin ());
             i != scope_stack.end ();
             ++i)
        {
          // We only qualify names until the namespace level.
          //
          if (first)
            first = false;
          else
            scope += L"::";

          scope += *i;
        }

        inner_name.clear ();

        if (!scope_stack.empty ())
          inner_name = scope_stack.back ();
      }

    public:
      CLI::Options const& options;
      Counts const& counts;
      String& parser_type;
      String& xml_element_type;
      String& xml_attribute_type;

      String& ns_scope;
      String& scope;
      String& inner_name;

    private:
      CLI::Options const options_;
      String parser_type_;
      String xml_element_type_;
      String xml_attribute_type_;

      typedef
      Cult::Containers::Deque<String>
      NamespaceStack;

      typedef
      Cult::Containers::Deque<String>
      ScopeStack;

      String ns_scope_;
      String scope_;
      String inner_name_;

      NamespaceStack& ns_scope_stack;
      NamespaceStack ns_scope_stack_;

      ScopeStack& scope_stack;
      ScopeStack scope_stack_;
    };

    // Check whether this Schema type maps to a fundamental C++ type.
    //
    struct IsFundamentalType : Traversal::Fundamental::Byte,
                               Traversal::Fundamental::UnsignedByte,
                               Traversal::Fundamental::Short,
                               Traversal::Fundamental::UnsignedShort,
                               Traversal::Fundamental::Int,
                               Traversal::Fundamental::UnsignedInt,
                               Traversal::Fundamental::Long,
                               Traversal::Fundamental::UnsignedLong,
                               Traversal::Fundamental::Integer,
                               Traversal::Fundamental::NonPositiveInteger,
                               Traversal::Fundamental::NonNegativeInteger,
                               Traversal::Fundamental::PositiveInteger,
                               Traversal::Fundamental::NegativeInteger,

                               Traversal::Fundamental::Boolean,

                               Traversal::Fundamental::Float,
                               Traversal::Fundamental::Double,
                               Traversal::Fundamental::Decimal

    {
      IsFundamentalType (Boolean& r)
          : r_ (r)
      {
      }

      // Integral types.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Byte&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedByte&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Short&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedShort&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Int&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedInt&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Long&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedLong&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Integer& i)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NonPositiveInteger& i)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NonNegativeInteger& i)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::PositiveInteger& i)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NegativeInteger& i)
      {
        r_ = true;
      }

      // Boolean.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Boolean&)
      {
        r_ = true;
      }

      // Floats.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Float&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Double&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Decimal& d)
      {
        r_ = true;
      }

    private:
      Boolean& r_;
    };

    // Check whether this is a string-based type.
    //
    struct IsStringBasedType : Traversal::Complex,
                               Traversal::Fundamental::String,
                               Traversal::Fundamental::NormalizedString,
                               Traversal::Fundamental::Token,
                               Traversal::Fundamental::Name,
                               Traversal::Fundamental::NameToken,
                               Traversal::Fundamental::NameTokens,
                               Traversal::Fundamental::NCName,
                               Traversal::Fundamental::Language
    {
      IsStringBasedType (Boolean& r)
          : r_ (r)
      {
        *this >> inherits_ >> *this;
      }

      virtual Void
      traverse (SemanticGraph::Complex& c)
      {
        inherits (c, inherits_);
      }

      // Strings.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::String& s)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NormalizedString& s)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Token& t)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NameToken& t)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NameTokens& t)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Name& n)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NCName& n)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Language& l)
      {
        r_ = true;
      }

    private:
      Boolean& r_;
      Traversal::Inherits inherits_;
    };

    // Check whether this is a enumeration-based type.
    //
    struct IsEnumBasedType : Traversal::Complex
    {
      IsEnumBasedType (SemanticGraph::Enumeration*& e)
          : enum_ (e)
      {
        *this >> inherits_;

        inherits_ >> *this;
        inherits_ >> enum_;
      }

      virtual Void
      traverse (SemanticGraph::Complex& c)
      {
        inherits (c, inherits_);
      }

    private:
      struct Enumeration: Traversal::Enumeration
      {
        Enumeration (SemanticGraph::Enumeration*& e)
            : e_ (e)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          if (e_ == 0)
            e_ = &e;
        }

      private:
        SemanticGraph::Enumeration*& e_;
      };


    private:
      Enumeration enum_;
      Traversal::Inherits inherits_;
    };


    //
    //
    struct MemberTypeName : protected virtual Context,
                            Traversal::Type,
                            Traversal::List,
                            Traversal::Union,
                            Traversal::Complex,

                            Traversal::AnyType,
                            Traversal::AnySimpleType,

                            Traversal::Fundamental::Byte,
                            Traversal::Fundamental::UnsignedByte,
                            Traversal::Fundamental::Short,
                            Traversal::Fundamental::UnsignedShort,
                            Traversal::Fundamental::Int,
                            Traversal::Fundamental::UnsignedInt,
                            Traversal::Fundamental::Long,
                            Traversal::Fundamental::UnsignedLong,
                            Traversal::Fundamental::Integer,
                            Traversal::Fundamental::NonPositiveInteger,
                            Traversal::Fundamental::NonNegativeInteger,
                            Traversal::Fundamental::PositiveInteger,
                            Traversal::Fundamental::NegativeInteger,

                            Traversal::Fundamental::Boolean,

                            Traversal::Fundamental::Float,
                            Traversal::Fundamental::Double,
                            Traversal::Fundamental::Decimal,

                            Traversal::Fundamental::String,
                            Traversal::Fundamental::NormalizedString,
                            Traversal::Fundamental::Token,
                            Traversal::Fundamental::Name,
                            Traversal::Fundamental::NameToken,
                            Traversal::Fundamental::NameTokens,
                            Traversal::Fundamental::NCName,
                            Traversal::Fundamental::Language,

                            Traversal::Fundamental::QName,

                            Traversal::Fundamental::Id,
                            Traversal::Fundamental::IdRef,
                            Traversal::Fundamental::IdRefs,

                            Traversal::Fundamental::AnyURI,

                            Traversal::Fundamental::Base64Binary,
                            Traversal::Fundamental::HexBinary,

                            Traversal::Fundamental::Date,
                            Traversal::Fundamental::DateTime,
                            Traversal::Fundamental::Duration,
                            Traversal::Fundamental::Day,
                            Traversal::Fundamental::Month,
                            Traversal::Fundamental::MonthDay,
                            Traversal::Fundamental::Year,
                            Traversal::Fundamental::YearMonth,
                            Traversal::Fundamental::Time,

                            Traversal::Fundamental::Entity,
                            Traversal::Fundamental::Entities

    {
      MemberTypeName (Context& c)
          : Context (c)
      {
      }

      MemberTypeName (Context& c, std::wostream& o)
          : Context (c, o)
      {
      }

      virtual Void
      traverse (SemanticGraph::Type& t)
      {
        abort ();
      }

      virtual Void
      traverse (SemanticGraph::List& l)
      {
        os << fq_name (l.scope ()) << "::" <<
          escape (L"_xsd_" + escape (l.name ())) << "::type";

        //else os << "class " << fq_name (l);
      }

      virtual Void
      traverse (SemanticGraph::Union& u)
      {
        os << fq_name (u.scope ()) << "::" <<
          escape (L"_xsd_" + escape (u.name ())) << "::type";

        //else os << "class " << fq_name (u);
      }

      virtual Void
      traverse (SemanticGraph::Complex& c)
      {
        os << fq_name (c.scope ()) << "::" <<
          escape (L"_xsd_" + escape (c.name ())) << "::type";
      }

      // anyType & anySimpleType.
      //
      virtual Void
      traverse (SemanticGraph::AnyType& i)
      {
        os << ns (i) << "::type";
      }

      virtual Void
      traverse (SemanticGraph::AnySimpleType& i)
      {
        os << ns (i) << "::simple_type";
      }

      // Integral types.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Byte&)
      {
        os << "signed char";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedByte&)
      {
        os << "unsigned char";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Short&)
      {
        os << "short";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedShort&)
      {
        os << "unsigned short";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Int&)
      {
        os << "int";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedInt&)
      {
        os << "unsigned int";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Long&)
      {
        os << "long long";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedLong&)
      {
        os << "unsigned long long";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Integer& i)
      {
        os << ns (i) << "::integer";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NonPositiveInteger& i)
      {
        os << ns (i) << "::non_positive_integer";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NonNegativeInteger& i)
      {
        os << ns (i) << "::non_negative_integer";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::PositiveInteger& i)
      {
        os << ns (i) << "::positive_integer";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NegativeInteger& i)
      {
        os << ns (i) << "::negative_integer";
      }

      // Boolean.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Boolean&)
      {
        os << "bool";
      }

      // Floats.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Float&)
      {
        os << "float";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Double&)
      {
        os << "double";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Decimal& d)
      {
        os << ns (d) << "::decimal";
      }

      // Strings.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::String& s)
      {
        os << ns (s) << "::string";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NormalizedString& s)
      {
        os << ns (s) << "::normalized_string";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Token& t)
      {
        os << ns (t) << "::token";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NameToken& t)
      {
        os << ns (t) << "::nmtoken";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NameTokens& t)
      {
        os << ns (t) << "::nmtokens";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Name& n)
      {
        os << ns (n) << "::name";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NCName& n)
      {
        os << ns (n) << "::ncname";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Language& l)
      {
        os << ns (l) << "::language";
      }


      // Qualified name.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::QName& n)
      {
        os << ns (n) << "::qname";
      }


      // ID/IDREF.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Id& i)
      {
        os << ns (i) << "::id";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::IdRef& i)
      {
        if (i.named ())
        {
          // IDREF<anyType>
          //
          os << ns (i) << "::idref";
        }
        else
        {
          os << "::xsd::cxx::tree::idref< " <<
            type_name (i.argumented ().type ()) << ", " << char_type << " >";
        }
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::IdRefs& i)
      {
        if (i.named ())
        {
          // IDREFS<anyType>
          //
          os << ns (i) << "::idrefs";
        }
        else
        {
          os << "::xsd::cxx::tree::idrefs< " <<
            type_name (i.argumented ().type ()) << ", " << char_type << " >";
        }
      }

      // URI.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::AnyURI& u)
      {
        os << ns (u) << "::uri";
      }

      // Binary.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Base64Binary& b)
      {
        os << ns (b) << "::base64_binary";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::HexBinary& b)
      {
        os << ns (b) << "::hex_binary";
      }


      // Date/time.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Date& d)
      {
        os << ns (d) << "::date";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::DateTime& dt)
      {
        os << ns (dt) << "::date_time";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Duration& d)
      {
        os << ns (d) << "::duration";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Day& d)
      {
        os << ns (d) << "::day";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Month& m)
      {
        os << ns (m) << "::month";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::MonthDay& md)
      {
        os << ns (md) << "::month_day";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Year& y)
      {
        os << ns (y) << "::year";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::YearMonth& ym)
      {
        os << ns (ym) << "::year_month";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Time& t)
      {
        os << ns (t) << "::time";
      }

      // Entity.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Entity& e)
      {
        os << ns (e) << "::entity";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Entities& e)
      {
        os << ns (e) << "::entities";
      }

    private:
      String
      ns (SemanticGraph::Type& t)
      {
        return ns_name (dynamic_cast<SemanticGraph::Namespace&> (t.scope ()));
      }

      // For idref/idrefs
      //
      String
      type_name (SemanticGraph::Type& t)
      {
        // This type is always named.
        //
        std::wostringstream o;

        MemberTypeName type (*this, o);
        type.dispatch (t);

        return o.str ();
      }
    };


    //
    //
    struct BaseTypeName : protected virtual Context,
                          Traversal::Type,
                          Traversal::List,
                          Traversal::Union,
                          Traversal::Complex,

                          Traversal::AnyType,
                          Traversal::AnySimpleType,

                          Traversal::Fundamental::Byte,
                          Traversal::Fundamental::UnsignedByte,
                          Traversal::Fundamental::Short,
                          Traversal::Fundamental::UnsignedShort,
                          Traversal::Fundamental::Int,
                          Traversal::Fundamental::UnsignedInt,
                          Traversal::Fundamental::Long,
                          Traversal::Fundamental::UnsignedLong,
                          Traversal::Fundamental::Integer,
                          Traversal::Fundamental::NonPositiveInteger,
                          Traversal::Fundamental::NonNegativeInteger,
                          Traversal::Fundamental::PositiveInteger,
                          Traversal::Fundamental::NegativeInteger,

                          Traversal::Fundamental::Boolean,

                          Traversal::Fundamental::Float,
                          Traversal::Fundamental::Double,
                          Traversal::Fundamental::Decimal,

                          Traversal::Fundamental::String,
                          Traversal::Fundamental::NormalizedString,
                          Traversal::Fundamental::Token,
                          Traversal::Fundamental::Name,
                          Traversal::Fundamental::NameToken,
                          Traversal::Fundamental::NameTokens,
                          Traversal::Fundamental::NCName,
                          Traversal::Fundamental::Language,

                          Traversal::Fundamental::QName,

                          Traversal::Fundamental::Id,
                          Traversal::Fundamental::IdRef,
                          Traversal::Fundamental::IdRefs,

                          Traversal::Fundamental::AnyURI,

                          Traversal::Fundamental::Base64Binary,
                          Traversal::Fundamental::HexBinary,

                          Traversal::Fundamental::Date,
                          Traversal::Fundamental::DateTime,
                          Traversal::Fundamental::Duration,
                          Traversal::Fundamental::Day,
                          Traversal::Fundamental::Month,
                          Traversal::Fundamental::MonthDay,
                          Traversal::Fundamental::Year,
                          Traversal::Fundamental::YearMonth,
                          Traversal::Fundamental::Time,

                          Traversal::Fundamental::Entity,
                          Traversal::Fundamental::Entities
    {
      BaseTypeName (Context& c, Boolean alias = true)
          : Context (c), alias_ (alias)
      {
      }

      BaseTypeName (Context& c, std::wostream& o, Boolean alias = true)
          : Context (c, o), alias_ (alias)
      {
      }

      virtual Void
      traverse (SemanticGraph::Type& t)
      {
        abort ();
      }

      virtual Void
      traverse (SemanticGraph::List& l)
      {
        if (alias_)
          os << fq_name (l.scope ()) << "::" <<
            escape (L"_xsd_" + escape (l.name ())) << "::type";
        else
          os << fq_name (l);
      }

      virtual Void
      traverse (SemanticGraph::Union& u)
      {
        if (alias_)
          os << fq_name (u.scope ()) << "::" <<
            escape (L"_xsd_" + escape (u.name ())) << "::type";
        else
          os << fq_name (u);
      }

      virtual Void
      traverse (SemanticGraph::Complex& c)
      {
        if (alias_)
          os << fq_name (c.scope ()) << "::" <<
            escape (L"_xsd_" + escape (c.name ())) << "::type";
        else
          os << fq_name (c);
      }

      // anyType & anySimpleType.
      //
      virtual Void
      traverse (SemanticGraph::AnyType& i)
      {
        os << ns (i) << "::type";
      }

      virtual Void
      traverse (SemanticGraph::AnySimpleType& i)
      {
        os << ns (i) << "::simple_type";
      }

      // Integrals.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Byte&)
      {
        os << "::xsd::cxx::tree::fundamental_base< signed char >";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedByte&)
      {
        os << "::xsd::cxx::tree::fundamental_base< unsigned char >";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Short&)
      {
        os << "::xsd::cxx::tree::fundamental_base< short >";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedShort&)
      {
        os << "::xsd::cxx::tree::fundamental_base< unsigned short >";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Int&)
      {
        os << "::xsd::cxx::tree::fundamental_base< int >";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedInt&)
      {
        os << "::xsd::cxx::tree::fundamental_base< unsigned int >";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Long&)
      {
        os << "::xsd::cxx::tree::fundamental_base< long long >";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedLong&)
      {
        os << "::xsd::cxx::tree::fundamental_base< unsigned long long >";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Integer& i)
      {
        os << "::xsd::cxx::tree::fundamental_base< " <<
          ns (i) << "::integer >";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NonPositiveInteger& i)
      {
        os << "::xsd::cxx::tree::fundamental_base< " <<
          ns (i) << "::non_positive_integer >";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NonNegativeInteger& i)
      {
        os << "::xsd::cxx::tree::fundamental_base< " <<
          ns (i) << "::non_negative_integer >";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::PositiveInteger& i)
      {
        os << "::xsd::cxx::tree::fundamental_base< " <<
          ns (i) << "::positive_integer >";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NegativeInteger& i)
      {
        os << "::xsd::cxx::tree::fundamental_base< " <<
          ns (i) << "::negative_integer >";
      }

      // Boolean.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Boolean&)
      {
        os << "::xsd::cxx::tree::fundamental_base< bool >";
      }

      // Floats.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Float&)
      {
        os << "::xsd::cxx::tree::fundamental_base< float >";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Double&)
      {
        os << "::xsd::cxx::tree::fundamental_base< double >";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Decimal& d)
      {
        os << "::xsd::cxx::tree::fundamental_base< " <<
          ns (d) << "::decimal >";
      }

      // Strings.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::String& s)
      {
        os << ns (s) << "::string";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NormalizedString& s)
      {
        os << ns (s) << "::normalized_string";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Token& t)
      {
        os << ns (t) << "::token";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NameToken& t)
      {
        os << ns (t) << "::nmtoken";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NameTokens& t)
      {
        os << ns (t) << "::nmtokens";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Name& n)
      {
        os << ns (n) << "::name";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NCName& n)
      {
        os << ns (n) << "::ncname";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Language& l)
      {
        os << ns (l) << "::language";
      }


      // Qualified name.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::QName& n)
      {
        os << ns (n) << "::qname";
      }


      // ID/IDREF.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Id& i)
      {
        os << ns (i) << "::id";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::IdRef& i)
      {
        if (i.named ())
        {
          // IDREF<anyType>
          //
          os << ns (i) << "::idref";
        }
        else
        {
          os << "::xsd::cxx::tree::idref< " <<
            type_name (i.argumented ().type ()) << ", " << char_type << " >";
        }
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::IdRefs& i)
      {
        if (i.named ())
        {
          // IDREF<anyType>
          //
          os << ns (i) << "::idrefs";
        }
        else
        {
          os << "::xsd::cxx::tree::idrefs< " <<
            type_name (i.argumented ().type ()) << ", " << char_type << " >";
        }
      }


      // URI.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::AnyURI& u)
      {
        os << ns (u) << "::id";
      }

      // Binary.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Base64Binary& b)
      {
        os << ns (b) << "::base64_binary";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::HexBinary& b)
      {
        os << ns (b) << "::hex_binary";
      }


      // Date/time.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Date& d)
      {
        os << ns (d) << "::date";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::DateTime& dt)
      {
        os << ns (dt) << "::date_time";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Duration& d)
      {
        os << ns (d) << "::duration";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Day& d)
      {
        os << ns (d) << "::day";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Month& m)
      {
        os << ns (m) << "::month";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::MonthDay& md)
      {
        os << ns (md) << "::month_day";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Year& y)
      {
        os << ns (y) << "::year";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::YearMonth& ym)
      {
        os << ns (ym) << "::year_month";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Time& t)
      {
        os << ns (t) << "::time";
      }

      // Entity.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Entity& e)
      {
        os << ns (e) << "::entity";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Entities& e)
      {
        os << ns (e) << "::entities";
      }

    private:
      String
      ns (SemanticGraph::Type& t)
      {
        return ns_name (dynamic_cast<SemanticGraph::Namespace&> (t.scope ()));
      }

      // For idref/idrefs.
      //
      String
      type_name (SemanticGraph::Type& t)
      {
        // This type is always named.
        //
        std::wostringstream o;

        MemberTypeName type (*this, o);
        type.dispatch (t);

        return o.str ();
      }

    private:
      Boolean alias_;
    };


    //
    //
    struct IsSimpleType : Traversal::Complex
    {
      IsSimpleType (Boolean& v)
          : v_ (v)
      {
        edge_traverser (inherits_);
        inherits_.node_traverser (*this);
      }

      virtual Void
      traverse (SemanticGraph::Complex& c)
      {
        if (has<Traversal::Member> (c))
          v_ = false;
        else
          inherits (c, *this);
      }

    private:
      Traversal::Inherits inherits_;
      Boolean& v_;
    };

    // Test whether we need to generate default c-tor.
    //
    struct GenerateDefaultCtor: Traversal::Complex,
                                Traversal::Enumeration,
                                Traversal::Type,
                                Traversal::Member,
                                protected virtual Context
    {
      // generate should initially be false.
      //
      GenerateDefaultCtor (Context& c, Boolean& generate);

      virtual Void
      traverse (SemanticGraph::Type& t);

      virtual Void
      traverse (SemanticGraph::Enumeration& e);

      virtual Void
      traverse (SemanticGraph::Member& m);

      using Complex::traverse;

    private:
      Boolean& generate_;

    private:
      Traversal::Inherits inherits_;
      Traversal::Names names_;
    };

    // Test whether we need to generate from-base c-tor.
    //
    struct GenerateFromBaseCtor: Traversal::Complex
    {
      // generate should initially be false.
      //
      GenerateFromBaseCtor (Context& c, Boolean& generate);

      Void
      traverse (SemanticGraph::Complex& c);

    private:
      struct Traverser: Traversal::Complex,
                        Traversal::Member,
                        protected virtual Context
      {
        Traverser (Context& c, Boolean& generate);

        virtual Void
        traverse (SemanticGraph::Member& m);

      private:
        Boolean& generate_;

      private:
        Traversal::Inherits inherits_;
        Traversal::Names names_;
      } traverser_;

      Traversal::Inherits inherits_;
    };


    // Immediate non-optional member.
    //
    struct FromBaseCtorArg : Traversal::Member, protected virtual Context
    {
      FromBaseCtorArg (Context& c, Boolean arg);

      virtual Void
      traverse (SemanticGraph::Member& m);

    private:
      Boolean arg_;
    };

    // List of all non-optional members
    //
    struct CtorArgs : Traversal::Complex,
                      Traversal::Enumeration,
                      Traversal::Type,
                      Traversal::Member,
                      protected virtual Context
    {
      CtorArgs (Context& c, String const& name, Boolean arg);

      virtual Void
      traverse (SemanticGraph::Type& t);

      virtual Void
      traverse (SemanticGraph::Enumeration& e);

      virtual Void
      traverse (SemanticGraph::Member& m);

    private:
      String
      comma ();

    private:
      String name_;
      Boolean arg_;
      Boolean first_;

    private:
      Traversal::Inherits inherits_;
      Traversal::Names names_;
    };


    //
    //
    struct GlobalElementBase
    {
      GlobalElementBase (Context& c)
          : ctx_ (c)
      {
      }

      Boolean
      generate_p (SemanticGraph::Element&);

      Boolean
      doc_root_p (SemanticGraph::Element&);

    private:
      Context& ctx_;
    };


    //
    //
    struct Namespace: CXX::Namespace,
                      GlobalElementBase,
                      CXX::Namespace::ScopeTracker
    {
      Namespace (Context&,
                 Boolean set_scope,
                 UnsignedLong first = 1,
                 UnsignedLong last = 0);

      virtual Void
      traverse (Type&);

    private:
      virtual Void
      enter (String const& name);

      virtual Void
      leave ();

    private:
      Context& ctx_;
      Boolean set_scope_;
      UnsignedLong first_;
      UnsignedLong last_;
      UnsignedLong count_;
    };
  }
}

#endif  // CXX_TREE_ELEMENTS_HXX
