// file      : processing/inheritance/processor.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <processing/inheritance/processor.hxx>

#include <elements.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <cult/containers/set.hxx>

#include <iostream>
using std::wcerr;
using std::endl;

namespace Processing
{
  using namespace Cult;

  namespace SemanticGraph = XSDFrontend::SemanticGraph;
  namespace Traversal = XSDFrontend::Traversal;

  typedef WideString String;

  namespace Inheritance
  {
    namespace
    {
      class Context
      {
      public:
        Context (SemanticGraph::Schema& schema_,
                 SemanticGraph::Path const& file)
            : schema_path_ (file),
              schema (schema_),
              schema_path (schema_path_)
        {
        }

      protected:
        Context (Context& c)
            : schema (c.schema),
              schema_path (c.schema_path)
        {
        }

      private:
        SemanticGraph::Path const schema_path_;

      public:
        SemanticGraph::Schema& schema;
        SemanticGraph::Path const& schema_path;
      };

      typedef Containers::Set<SemanticGraph::Type*> TypeSet;


      // Calculate the list of dependencies for this complex
      // type.
      //
      struct ComplexType: Traversal::Complex,
                          Traversal::Member
      {
        ComplexType (TypeSet& dep_set)
            : dep_set_ (dep_set)
        {
          *this >> names_ >> *this;
        }

        virtual Void
        traverse (SemanticGraph::Complex& c)
        {
          using SemanticGraph::Complex;

          for (Complex::InheritsIterator i (c.inherits_begin ());
               i != c.inherits_end (); ++i)
            dep_set_.insert (&i->inheritee ());

          // Go after anonymous types.
          //
          names (c);
        }

        virtual Void
        traverse (SemanticGraph::Member& m)
        {
          SemanticGraph::Type& t (m.type ());

          if (!t.named () && !t.context ().count ("anon-seen"))
          {
            t.context ().set ("anon-seen", true);
            dispatch (t);
            t.context ().remove ("anon-seen");
          }
        }

      private:
        Traversal::Names names_;
        TypeSet& dep_set_;
      };


      //
      //
      template <typename N, typename A>
      struct NodeArgs
      {
        NodeArgs (N& node, A arg)
            : node_ (node), arg_ (arg)
        {
        }

        operator N& () const
        {
          return node_;
        }

        template <typename E>
        Void
        add_edge_left (E& e)
        {
          node_.add_edge_left (e, arg_);
        }

        template <typename E>
        Void
        add_edge_right (E& e)
        {
          node_.add_edge_right (e, arg_);
        }

      private:
        N& node_;
        A arg_;
      };


      //
      //
      struct Global: Traversal::Type,
                     Traversal::Complex,
                     Traversal::Element
      {
        Global (SemanticGraph::Schema& schema)
            : schema_ (schema)
        {
        }

        virtual Void
        traverse (SemanticGraph::Type& t)
        {
          if (t.named ())
            seen_types_.insert (&t);
        }

        virtual Void
        traverse (SemanticGraph::Complex& c)
        {
          check_dep (c, c);
          seen_types_.insert (&c);
        };

        virtual Void
        traverse (SemanticGraph::Element& e)
        {
          SemanticGraph::Type& t (e.type ());

          if (!t.named ())
            check_dep (e, t);
        };

      private:
        Void
        check_dep (SemanticGraph::Nameable& global,
                   SemanticGraph::Type& type)
        {
          using SemanticGraph::Type;
          using SemanticGraph::Scope;
          using SemanticGraph::Names;
          using SemanticGraph::Schema;

          TypeSet prereqs;

          // Calculate our prerequisistes.
          //
          {
            ComplexType complex (prereqs);
            complex.dispatch (type);
          }

          for (TypeSet::Iterator i (prereqs.begin ());
               i != prereqs.end (); ++i)
          {
            Type& t (**i);

            if (seen_types_.find (&t) == seen_types_.end ())
            {
              Scope& scope (t.scope ());
              Schema& schema (dynamic_cast<Schema&> (scope.scope ()));

              // Don't worry about types that are in included/imported
              // schemas.
              //
              if (&schema != &schema_ && !sources_p (schema_, schema))
                continue;

              if (t.context ().count ("seen"))
              {
                /*
                wcerr << "pathological case detected: '" << t.name ()
                      << "' cyclicly depends on '" << global.name () << "'"
                      << endl;
                */
                continue;
              }

              /*
              wcerr << "type '" << t.name () << "' needs to be moved " <<
                "before " << (global.is_a<Type> () ? "type" : "element") <<
                " '" << global.name () << "'" << endl;
              */

              // Delete current Names edge.
              //
              String name (t.name ());
              {
                Names& n (t.named_ ());
                schema_.delete_edge (scope, t, n);
              }

              // Insert a new Names edge before global.
              //
              {
                // Convert to the insert-after call.
                //
                Scope::NamesIterator i (scope.find (global.named_ ()));

                if (i == scope.names_begin ())
                  i = scope.names_end ();
                else
                  --i;

                NodeArgs<Scope, Scope::NamesIterator> na (scope, i);
                schema_.new_edge<Names> (na, t, name);
              }

              // Recursively process the moved type.
              //
              global.context ().set ("seen", true);
              dispatch (t);
              global.context ().remove ("seen");
            }
          }
        }


      private:
        // Return true if root sources s.
        //
        Boolean
        sources_p (SemanticGraph::Schema& root, SemanticGraph::Schema& s)
        {
          using SemanticGraph::Schema;
          using SemanticGraph::Sources;

          for (Schema::UsesIterator i (root.uses_begin ());
               i != root.uses_end (); ++i)
          {
            if (i->is_a<Sources> ())
            {
              if (&i->schema () == &s || sources_p (i->schema (), s))
                return true;
            }
          }

          return false;
        }

      private:
        SemanticGraph::Schema& schema_;
        TypeSet seen_types_;
      };
    }

    Void Processor::
    process (SemanticGraph::Schema& tu,
             SemanticGraph::Path const& file)
    {
      Traversal::Schema schema;
      Traversal::Sources sources;

      schema >> sources >> schema;

      Traversal::Names schema_names;
      Traversal::Namespace ns;
      Traversal::Names ns_names;

      schema >> schema_names >> ns >> ns_names;

      Global global (tu);

      ns_names >> global;

      schema.dispatch (tu);
    }
  }
}
